import sys
import traceback
from atelier_fabrique.message import Message, MessageType
from .logger import Logger

METRIC_FIELDS = ['error', 'alerts', 'metrics', 'raw_metrics', 'raw_alerts'] #hardcoded body fields for metrics and alerts

class MessageBus(object):
    def __init__(self, config_dict):
        
        
        self.metric_fields = METRIC_FIELDS 

        self.config = config_dict
        
        self.store = self.config['message_bus']['servers']

        self.client_id = self.config["message_bus"]["client_id"] 
        
        self.pipeline_ver = self.config["message_bus"].get("pipeline_ver") 
        self.pipeline_name = self.config["message_bus"].get("pipeline_name")

        #logger
        self.logger = Logger.make_logger(self.client_id)
        
        #producer+consumer                 
        to_list = lambda x: x if type(x)==list else ([x,] if x else []) #converts all types to list type        
        #1. лист входных топиков или один топик - неважно
        self.in_topics = to_list(self.config['message_bus'].get('in_topic')) 
        #2. вЫходной топик для данных, результатов, скоров и т.д.
        self.out_topic = self.config['message_bus'].get('out_topic') 
        #3. топик для ошибок
        self.error_topic = self.config['message_bus'].get('error_topic') 
        #4. топик для мониторинга
        self.monitor_topics = self.config['message_bus'].get('monitor_topics')
        #5. выходные топики для диспетчера
        self.out_topic_dict = self.config['message_bus'].get('out_topic_dict')
    
    def send_message(self, body, session_id=None, 
                                 message_generator=Message.gen_data_message, 
                                 topic=None):
        
        self.logger.debug('! send msg with session_id = {} to topic = {}'.format(session_id, topic))
            
        if not topic: #для тестовых пайплайнов нормально не иметь выходного топика
            return 
        
        mes = message_generator(session_id, self.client_id, body, pipeline_name=self.pipeline_name, 
                                                          pipeline_ver=self.pipeline_ver)
        try:
            self.store.produce_kaf_format(topic, mes.serialize())
        except TypeError:
            print(f'!Ffffffuuuuuuu, cannot send to {topic}, {type(self).__name__}, {mes.body}')
            raise
        
    def send_data(self, body, session_id, topic=None, message_generator=Message.gen_data_message):
        '''Отправлялка данных'''
        if not topic:
            topic=self.out_topic
        
        if topic: #for metric_reducer is empty
            self.send_message(dict(data=body['data']), session_id, 
                          message_generator=message_generator, topic=topic)
        
        self.send_monitor_metric(body, session_id)
        
        
    def send_error(self, body, session_id, topic=None):
        '''Отправлялка ошибок'''
        if not topic:
            topic = self.error_topic
        if not topic:
            return
        
        message_generator = Message.gen_error_message
        
        self.send_message(dict(error=body['error']), 
                          session_id, message_generator=message_generator, topic=topic)
        
        self.send_monitor_metric(body, session_id)                                 
                        
    def send_monitor_metric(self, body, session_id, topics=[]):
        '''Отправлялка в мониторинг'''
        
        if not topics:
            topics = self.monitor_topics
        
        #монторинг понимает только ограниченный набор полей на метрики
        metric_body = {field:body[field] for field in self.metric_fields if field in body}
            
        if metric_body:
            message_generator = Message.gen_data_message
            
            for topic in topics:
                self.logger.info(f'! send metric with session_id = {session_id} to topic = {topic}: {metric_body}')
                self.send_message(metric_body, session_id, 
                              message_generator=message_generator, topic=topic)
       
                         
class Actor(MessageBus):
    
    def preprocess_msg(self, msg):        
        '''пробуем распарсить
           если дадим на выход None, в колбек ничего не попадет'''
        
        try:            
            message = Message.parse_mes(msg)
        except Exception as e:
            msg = traceback.format_exc()
            error = '! Cannot parse message in {0!s}: {1!r}'.format(type(self).__name__, msg)
            self.logger.error(error)
            self.send_error({'error':error}, 'not_specified') #сессию определить не удалось, но матюгнуться стоит!
            raise
            
            return None 
                         
        if message.header.message_type != MessageType.DATA.value:
            return None
        
        return message
                         
    def batch_callback(self, bodies):
        #в наследуемом классе, body - тело месседжа, на выходе res_body
        raise Exception('batch_callback method is not defined')
        
    def close_all(self):
        self.store.stop = True
        
    def run(self):
        self.logger.info('{} started'.format(self.client_id))
        try:              
            for messages in self.store.reader_generator(self.in_topics):
                errors_batch = []
                res_batch = []

                messages_batch = [[message.header.session_id, message.body] for message in 
                                  (self.preprocess_msg(mes) for mes in messages) if message]
                
                if not messages_batch:
                    continue
                        
                try:
                    session_ids, bodies = list(zip(*messages_batch))
                    out_bodies = self.batch_callback(bodies)
                    if not out_bodies:
                        continue
                        
                    assert(len(out_bodies)==len(bodies))
                        
                    for i, body in enumerate(out_bodies):
                        session_id = session_ids[i]
                        if 'error' in body:
                            errors_batch.append([session_id, body])
                        else:
                            res_batch.append([session_id, body])
                                
                except Exception as e:
                    msg = traceback.format_exc()
                    error = "Exception in {0!s}: {1!r}".format(type(self).__name__, msg)
                    self.logger.error(error)
                    errors_batch = [[session_id, {'error':error}] for session_id, body in  messages_batch]
                    raise
                    
                for session_id, body in res_batch:                
                    self.process_result(body, session_id)
                        
                for session_id, body in errors_batch:
                    self.send_error(body, session_id)

        except KeyboardInterrupt:
            pass
                         
        except Exception as e:
            trace_msg = traceback.format_exc()
            self.logger.error("Exception: {0!r}".format(trace_msg))
            raise
            
        finally:
            self.close_all()

    def check_key_presence_in_res_body(self, res_body, session_id, key):
        if key in res_body:
            return True        
        error = "Exception in {}: No '{}' key in res_body".format(type(self).__name__, key)
        self.logger.error(error)
        self.send_error({'error':error}, session_id)
    
        
class Reducer(Actor):        
    def __init__(self, config_dict):
        super().__init__(config_dict) 

    def process_result(self, res_body, session_id):
        self.send_data(res_body, session_id)
        
class Mapper(Actor):        
    def __init__(self, config_dict):
        super().__init__(config_dict) 

    def process_result(self, res_body, session_id):
        self.send_data(res_body, session_id)

    
        
class Dispatcher(Actor):        
    """Dispatcher умеет читать поле dispatch и отправлять сообщение в нужный id по топику
    """    
    def __init__(self, config_dict):
        super().__init__(config_dict) 

    def process_result(self, res_body, session_id):
        dest_id = res_body.pop('destination', None)
        
        if not dest_id:
            error = "Exception in {0!s}: {1}".format(type(self).__name__, 'no "destination" field in result')            
            self.send_error({'error':error}, session_id)
            return
        
        try:
            topic = self.out_topic_dict[dest_id]
            
        except:
            error = 'Exception in {0!s}: no topic in topic_dict for destination "{1}"'.format(type(self).__name__, dest_id)
            self.send_error({'error':error}, session_id)
            return
        
        self.send_data(res_body, session_id, topic=topic)
        
