from enum import Enum, unique
from copy import deepcopy
import msgpack, json
import uuid

@unique
class MessageType(Enum):
    ERROR = 'ERROR' #ошибка
    DATA = 'DATA' #data 
    MANAGEMENT = 'MANAGEMENT' #указания от pipeline manager

class Header(object):
    def __init__(self, session_id, 
                       message_type,
                       source,
                       corr_id=None,
                       pipeline_name=None,
                       pipeline_ver=None):
        
        if not corr_id:
            corr_id = str(uuid.uuid4())
            
        self.corr_id = corr_id
        self.session_id = session_id
        self.pipeline_ver = pipeline_ver
        self.pipeline_name = pipeline_name
        self.message_type = message_type
        self.source = source
        
    def to_dict(self):
        return dict(session_id=self.session_id,
                    message_type=self.message_type,
                    source=self.source,
                    corr_id=self.corr_id,
                    pipeline_name=self.pipeline_name,
                    pipeline_ver=self.pipeline_ver,
                   )


class Message(object):
    '''класс должен
    1. задавать структуру
    2. парсить месседж кафки
    3. генерить на основе принятого заголовка новый заголовок
    4. генерить типовые заголовки для резалта, ошибки, мониторинга и т.д.
    5. сериализовывать перед отправкой
    + добавил корявенько имя и версию пайплайна'''
    
    def __init__(self, header, body, kafka_header={}):
        self.kafka_header = kafka_header
        self.header = header
        self.body = body     
    
    def serialize(self):
        header_dict = self.header.to_dict()
        val_dict = dict(header=header_dict, body=self.body)
        
        packed = msgpack.dumps(val_dict, use_bin_type=True) #json.dumps(val_dict).encode()
        return packed
    
    @classmethod
    def parse_mes(cls, mes):
        #создавалка месседжа из сообщения кафки
        kafka_header = dict(topic=mes.topic, partition=mes.partition, offset=mes.offset, 
                            timestamp=mes.timestamp, key=mes.key)
        value = mes.value
        #val_dict = json.loads(value.decode())
        val_dict = msgpack.loads(value, encoding='utf-8')
        header_dict = val_dict['header']
        body = val_dict['body']
        header = Header(header_dict['session_id'], 
                        header_dict['message_type'], 
                        header_dict['source'], 
                        header_dict['corr_id'], 
                        header_dict.get('pipeline_name'),
                        header_dict.get('pipeline_ver'))
        return cls(header, body, kafka_header) 
    
    @classmethod
    def gen_message(cls, session_id, source, body, message_type, pipeline_name=None, pipeline_ver=None):
        header = Header(session_id, 
                        message_type,
                        source, pipeline_name=pipeline_name, pipeline_ver=pipeline_ver)
        return cls(header, body)
    
    @classmethod
    def gen_error_message(cls, session_id, source, body, pipeline_name=None, pipeline_ver=None):
        return cls.gen_message(session_id, source, body, MessageType.ERROR.value, 
                               pipeline_name=pipeline_name, pipeline_ver=pipeline_ver)
    
    @classmethod
    def gen_data_message(cls, session_id, source, body, pipeline_name=None, pipeline_ver=None):
        return cls.gen_message(session_id, source, body, MessageType.DATA.value, 
                               pipeline_name=pipeline_name, pipeline_ver=pipeline_ver) 
    
    @classmethod
    def gen_management_message(cls, session_id, source, body, pipeline_name=None, pipeline_ver=None):
        return cls.gen_message(session_id, source, body, MessageType.MANAGEMENT.value, 
                               pipeline_name=pipeline_name, pipeline_ver=pipeline_ver)  
